#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "packet.h"
#include "scan.h"

void create_udp_packet(struct udphdr *udphdr, const struct scan *data)
{
	udphdr->source = htons(1234);
	udphdr->dest = htons(data->port);
	udphdr->len = sizeof(struct udphdr);
	udphdr->check = checksum(udphdr, sizeof(struct udphdr));
}

int create_tcp_packet(struct tcphdr *tcphdr, const struct scan *data)
{
	tcphdr->source = htons(1234);
	tcphdr->dest = htons(data->port);
	tcphdr->doff = sizeof(struct tcphdr) / sizeof(int);
	tcphdr->fin = data->type == SCAN_XMAS || data->type == SCAN_FIN;
	tcphdr->syn = data->type == SCAN_SYN;
	tcphdr->psh = data->type == SCAN_XMAS;
	tcphdr->ack = data->type == SCAN_ACK;
	tcphdr->urg = data->type == SCAN_XMAS;
	tcphdr->window = htons(5840);
	tcphdr->urg_ptr = 0;

	struct pshdr pshdr;
	pshdr.src_addr = inet_addr(data->host->ip);
	pshdr.dst_addr = inet_addr(data->dest_addr);
	pshdr.placeholder = 0;
	pshdr.protocol = IPPROTO_TCP;
	pshdr.tcp_length = htons(sizeof(struct tcphdr));

	size_t pseudogram_size = sizeof(struct pshdr) + sizeof(struct tcphdr);
	void *pseudogram = malloc(pseudogram_size);
	if (!pseudogram)
		return -1;
	memcpy(pseudogram, &pshdr, sizeof(pshdr));
	memcpy(pseudogram + sizeof(pshdr), tcphdr, sizeof(struct tcphdr));
	tcphdr->check = checksum(pseudogram, pseudogram_size);
	free(pseudogram);

	return 0;
}

static void *create_packet(const struct scan *data, size_t packet_size)
{
	const bool isudp = data->type == SCAN_UDP;
	void *packet = calloc(packet_size, 1);
	if (!packet) {
		dprintf(2,
		        "ft_nmap: allocation failed during packet creation\n");
		return NULL;
	}

	struct iphdr *iphdr = packet;
	iphdr->ihl = sizeof(struct iphdr) / sizeof(int);
	iphdr->version = 4;
	iphdr->tos = 0;
	iphdr->tot_len = packet_size;
	iphdr->id = htons(54321);
	iphdr->frag_off = 0;
	iphdr->ttl = data->ttl;
	iphdr->protocol = isudp ? IPPROTO_UDP : IPPROTO_TCP;
	iphdr->saddr = inet_addr(data->host->ip);
	iphdr->daddr = inet_addr(data->dest_addr);

	iphdr->check = checksum(packet, packet_size);

	// this is starnakin stuff
	switch ((int)isudp) {
	case true:
		create_udp_packet(packet + sizeof(struct iphdr), data);
		break;
	default:
		create_tcp_packet(packet + sizeof(struct iphdr), data);
	}

	return packet;
}

int send_packet(const struct scan *data, int sockfd)
{
	struct sockaddr_in conn_addr;
	conn_addr.sin_family = AF_INET;
	conn_addr.sin_port = htons(80);
	conn_addr.sin_addr.s_addr = inet_addr(data->dest_addr);

	size_t packet_size = sizeof(struct iphdr) +
	                     (data->type == SCAN_UDP ? sizeof(struct udphdr)
	                                             : sizeof(struct tcphdr));
	void *packet = create_packet(data, packet_size);

	int one = 1;
	const int *val = &one;
	if (setsockopt(sockfd, IPPROTO_IP, IP_HDRINCL, val, sizeof(one)) < 0) {
		perror("Error setting IP_HDRINCL");
		return -1;
	}

	sendto(sockfd, packet, packet_size, 0, (struct sockaddr *)&conn_addr,
	       sizeof(struct sockaddr_in));
	free(packet);

	return 0;
}
